#include <stdio.h>
#include <stdlib.h>

#include "el.h"
#include "config.h"
#include "md5.h"

enum md5userdefined
  {
    I0_MD5_CTX
  };

struct md5storage
{
  MD5_CTX md5;		/* the context holder */
};

int
md5__init__t (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  struct md5storage *md5p;

  md5p = (struct md5storage *) malloc (sizeof (struct md5storage));
  MD5Init (&md5p->md5);

  RvSetUserData (self, I0_MD5_CTX, (void *) md5p);

  return 0;
}

int
md5__init__ts (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  int len = 0;
  unsigned char *cp = NULL;
  struct md5storage *md5p;

  if (RvParseArg (self, args, "s#", &cp, &len))
    return -1;

  md5p = (struct md5storage *) malloc (sizeof (struct md5storage));
  MD5Init (&md5p->md5);

  if (cp)
    MD5Update (&md5p->md5, cp, len);

  RvSetUserData (self, I0_MD5_CTX, (void *) md5p);

  return 0;
}

int
md5__del__ (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  struct md5storage *md5p;

  md5p = (struct md5storage *) RvGetUserData (self, I0_MD5_CTX);
  if (!md5p)
    return 0;

  free (md5p);
  RvSetUserData (self, I0_MD5_CTX, NULL);

  return 0;
}

int
md5_update_ts (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  int len = 0;
  unsigned char *cp = NULL;
  struct md5storage *md5p;

  if (RvParseArg (self, args, "s#", &cp, &len))
    return -1;

  md5p = (struct md5storage *) RvGetUserData (self, I0_MD5_CTX);
  if (!md5p)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "Initialize class first.");
      return -1;
    }

  MD5Update (&md5p->md5, cp, len);

  return 0;
}

int
md5_digest_t (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  MD5_CTX mdContext;
  unsigned char aDigest[16];
  struct md5storage *md5p;

  md5p = (struct md5storage *) RvGetUserData (self, I0_MD5_CTX);
  if (!md5p)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "Initialize class first.");
      return -1;
    }

  /* make a temporary copy, and perform the final */
  mdContext = md5p->md5;
  MD5Final (aDigest, &mdContext);

  RvSetReturn (self, ret, "S", RvStrNew (self, (char *) aDigest, 16));

  return 0;
}

int
md5_hexdigest_t (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  MD5_CTX mdContext;
  unsigned char digest[16];
  unsigned char hexdigest[32];
  int i, j;
  struct md5storage *md5p;
  
  md5p = (struct md5storage *) RvGetUserData (self, I0_MD5_CTX);
  if (!md5p)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "Initialize class first.");
      return -1;
    }
  
  /* make a temporary copy, and perform the final */
  mdContext = md5p->md5;
  MD5Final (digest, &mdContext);
  
  /* Make hex version of the digest */
  for(i=j=0; i<16; i++) 
    {
      char c;
      c = (digest[i] >> 4) & 0xf;
      c = (c>9) ? c+'a'-10 : c + '0';
      hexdigest[j++] = c;
      c = (digest[i] & 0xf);
      c = (c>9) ? c+'a'-10 : c + '0';
      hexdigest[j++] = c;
    }

  RvSetReturn (self, ret, "S", RvStrNew (self, (char*) hexdigest, 32));

  return 0;
}

int
md5_copy_t (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  RvObject *md5;
  struct md5storage *md5p, *clone;
  
  md5p = (struct md5storage *) RvGetUserData (self, I0_MD5_CTX);
  if (!md5p)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "Initialize class first.");
      return -1;
    }

  md5 = RvObjectNew (self, 0, "localhost", 0, "/test/md5");
  if (md5__init__t (md5, args, ret))
    return -1;

  clone = (struct md5storage *) RvGetUserData (md5, I0_MD5_CTX);
  clone->md5 = md5p->md5;

  RvSetReturn (self, ret, "T", md5);

  return 0;
}

static RvMethodDef method[] =
  {
    { "__init__",       "(T)V",         md5__init__t },
    { "__init__",       "(TS)V",        md5__init__ts },
    { "__del__",        "(T)V",         md5__del__ },
    { "update",         "(TS)V",        md5_update_ts },
    { "digest",         "(T)S",         md5_digest_t },
    { "hexdigest",      "(T)S",         md5_hexdigest_t },
    { "copy",           "(T)T",         md5_copy_t },
    { NULL, },
  };

/** entry point.  */
int
init_md5 (RvClass *cls)
{
  if (RvDefineMethods (cls, method))
    return -1;

  return 0;
}
