#include <stdio.h>
#include <stdlib.h>

#include "el.h"
#define ROVM
#include "regex.h"

#include "regexmodule.h"

static RvFieldDef field[] = 
  {
    { "lastok",         "S" },
    { "groupindex",     "T" },
    { "givenpat",       "S" },
    { "realpat",        "S" },
    { NULL, NULL },
  };

int
rr__del__ (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  rovmregex_t *rt;

  rt = (rovmregex_t *) RvGetUserData (self, UD_RE);
  if (!rt)
    return 0;

  if (rt->regs)
    re_free_registers (rt->regs);

  re_free_pattern (rt->re_patbuf);

  free (rt);

  return 0;
}

int
rr_match_ts (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  int offset = 0, size, result;
  char *buffer;
  rovmregex_t *rt;
  RvObject *argstring;

  rt = (rovmregex_t *) RvGetUserData (self, UD_RE);
  if (!rt)
    return 0;

  if (RvParseArg (self, args, "S", &argstring))
    return -1;

  if (RvStrGetInfo (argstring, &buffer, &size))
    return -1;

  if (offset < 0 || offset > size)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "match offset out of range");
      return -1;
    }

  RvSetField (self, "lastok", "S", NULL_OBJECT);

  rt->regs = &rt->re_regs;
  result = rovm_re_match (rt->re_patbuf, (unsigned char *) buffer, size, offset, &rt->re_regs);
  if (result < -1)
    {
      /* Serious failure of some sort; if re_match didn't 
         set an exception, raise a generic error */
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "match failure");
      return -1;
    }

  if (result >= 0)
    RvSetField (self, "lastok", "S", argstring);

  RvSetReturn (self, ret, "I", result); /* Length of the match or -1 */

  return 0;
}

int
rr_match_tsi (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  int offset = 0, size, result;
  char *buffer;
  rovmregex_t *rt;
  RvObject *argstring;

  rt = (rovmregex_t *) RvGetUserData (self, UD_RE);
  if (!rt)
    return 0;

  if (RvParseArg (self, args, "Si", &argstring, &offset))
    return -1;

  if (RvStrGetInfo (argstring, &buffer, &size))
    return -1;

  if (offset < 0 || offset > size)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "match offset out of range");
      return -1;
    }

  RvSetField (self, "lastok", "S", NULL_OBJECT);

  rt->regs = &rt->re_regs;
  result = rovm_re_match (rt->re_patbuf, (unsigned char *) buffer, size, offset, &rt->re_regs);
  if (result < -1)
    {
      /* Serious failure of some sort; if re_match didn't 
         set an exception, raise a generic error */
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "match failure");
      return -1;
    }

  if (result >= 0)
    RvSetField (self, "lastok", "S", argstring);

  RvSetReturn (self, ret, "I", result); /* Length of the match or -1 */

  return 0;
}

int
rr_search_ts (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  int offset = 0, size, result, range;
  char *buffer;
  rovmregex_t *rt;
  RvObject *argstring;

  rt = (rovmregex_t *) RvGetUserData (self, UD_RE);
  if (!rt)
    return 0;

  if (RvParseArg (self, args, "S", &argstring))
    return -1;

  if (RvStrGetInfo (argstring, &buffer, &size))
    return -1;

  if (offset < 0 || offset > size)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "search offset out of range");
      return -1;
    }

  range = size - offset;

  RvSetField (self, "lastok", "S", NULL_OBJECT);

  rt->regs = &rt->re_regs;
  result = rovm_re_search (rt->re_patbuf, (unsigned char *) buffer, size, offset, range, &rt->re_regs);
  if (result < -1)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "search failure");
      return -1;
    }

  if (result >= 0)
    RvSetField (self, "lastok", "S", argstring);

  RvSetReturn (self, ret, "I", result); /* Length of the match or -1 */

  return 0;
}

int
rr_search_tsi (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  int offset = 0, size, result, range;
  char *buffer;
  rovmregex_t *rt;
  RvObject *argstring;

  rt = (rovmregex_t *) RvGetUserData (self, UD_RE);
  if (!rt)
    return 0;

  if (RvParseArg (self, args, "Si", &argstring, &offset))
    return -1;

  if (RvStrGetInfo (argstring, &buffer, &size))
    return -1;

  if (offset < 0 || offset > size)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "search offset out of range");
      return -1;
    }

  range = size - offset;

  RvSetField (self, "lastok", "S", NULL_OBJECT);

  rt->regs = &rt->re_regs;
  result = rovm_re_search (rt->re_patbuf, (unsigned char *) buffer, size, offset, range, &rt->re_regs);
  if (result < -1)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "search failure");
      return -1;
    }

  if (result >= 0)
    RvSetField (self, "lastok", "S", argstring);

  RvSetReturn (self, ret, "I", result); /* Length of the match or -1 */

  return 0;
}

int
rr_group_t (self, args, ret)
     RvObject *self;
     RvValue *args, *ret;
{
  int a, b, index, len;
  char *str;
  rovmregex_t *rt;
  RvObject *lastok, *result;

  rt = (rovmregex_t *) RvGetUserData (self, UD_RE);
  if (!rt)
    return 0;

  if (RvParseArg (self, args, "i", &index))
    return -1;

  if (index < 0 || index >= RE_NREGS)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, "group() index out of range");
      return -1;
    }

  if (RvGetField (self, "lastok", "S", &lastok))
    return -1;

  if (!lastok)
    {
      RvLogError (self, ERRLOG_ERR, ERRLOG_MARK, 
                  "group() only valid after successful match/search");
      return -1;
    }

  result = RvArrayNew (self, 0);

  a = rt->re_regs.beg[index];
  b = rt->re_regs.end[index];
  if (a < 0 || b < 0) 
    {
      RvSetReturn (self, ret, "[", result);
      return -1;
    }

  if (RvStrGetInfo (lastok, &str, &len))
    return -1;
  
  RvArrayPush (result, "S", RvStrNew (self, str + a, b - a));

  RvSetReturn (self, ret, "[", result);

  return 0;
}

static RvMethodDef method[] =
  {
    { "__del__",        "(T)V",         rr__del__ },
    { "match",          "(TS)I",        rr_match_ts },
    { "match",          "(TSI)I",       rr_match_tsi },
    { "search",         "(TS)I",        rr_search_ts },
    { "search",         "(TSI)I",       rr_search_tsi },
    { "group",          "(TI)[",        rr_group_t },
    { NULL, },
  };

int
init_regex_regex (RvClass *cls)
{
  if (RvDefineFields (cls, field))
    return -1;

  if (RvDefineMethods (cls, method))
    return -1;  

  return 0;
}
